@@ -190,6 +190,8 @@ static void profile_end(const char *f, int line)
 /** used for backward compatibility */
 #define SCARD_PROTOCOL_ANY_OLD	 0x1000
 
+static pthread_mutex_t LockMutex = PTHREAD_MUTEX_INITIALIZER;
+
 LONG SCardEstablishContext(DWORD dwScope, /*@unused@*/ LPCVOID pvReserved1,
 	/*@unused@*/ LPCVOID pvReserved2, LPSCARDCONTEXT phContext)
 {
@@ -232,7 +234,6 @@ LONG SCardConnect(/*@unused@*/ SCARDCONTEXT hContext, LPCSTR szReader,
 {
 	LONG rv;
 	READER_CONTEXT * rContext = NULL;
-	uint32_t readerState;
 
 	(void)hContext;
 	PROFILE_START
@@ -301,11 +302,9 @@ LONG SCardConnect(/*@unused@*/ SCARDCONTEXT hContext, LPCSTR szReader,
 	 * presence of a card or not
 	 *
 	 *******************************************/
-	readerState = rContext->readerState->readerState;
-
 	if (dwShareMode != SCARD_SHARE_DIRECT)
 	{
-		if (!(readerState & SCARD_PRESENT))
+		if (!(rContext->readerState->readerState & SCARD_PRESENT))
 		{
 			Log1(PCSC_LOG_DEBUG, "Card Not Inserted");
 			rv = SCARD_E_NO_SMARTCARD;
@@ -325,7 +324,7 @@ LONG SCardConnect(/*@unused@*/ SCARDCONTEXT hContext, LPCSTR szReader,
 
 			if (rv == IFD_SUCCESS)
 			{
-				readerState = SCARD_PRESENT | SCARD_POWERED | SCARD_NEGOTIABLE;
+				rContext->readerState->readerState = SCARD_PRESENT | SCARD_POWERED | SCARD_NEGOTIABLE;
 
 				Log1(PCSC_LOG_DEBUG, "power up complete.");
 				LogXxd(PCSC_LOG_DEBUG, "Card ATR: ",
@@ -337,7 +336,7 @@ LONG SCardConnect(/*@unused@*/ SCARDCONTEXT hContext, LPCSTR szReader,
 					rv, rv);
 		}
 
-		if (! (readerState & SCARD_POWERED))
+		if (! (rContext->readerState->readerState & SCARD_POWERED))
 		{
 			Log1(PCSC_LOG_ERROR, "Card Not Powered");
 			(void)pthread_mutex_unlock(&rContext->powerState_lock);
@@ -444,6 +443,11 @@ LONG SCardConnect(/*@unused@*/ SCARDCONTEXT hContext, LPCSTR szReader,
 	/*
 	 * Prepare the SCARDHANDLE identity
 	 */
+
+	/* we need a lock to avoid concurent generation of handles leading
+	 * to a possible hCard handle duplication */
+	(void)pthread_mutex_lock(&LockMutex);
+
 	*phCard = RFCreateReaderHandle(rContext);
 
 	Log2(PCSC_LOG_DEBUG, "hCard Identity: %lx", *phCard);
@@ -467,6 +471,7 @@ LONG SCardConnect(/*@unused@*/ SCARDCONTEXT hContext, LPCSTR szReader,
 			(void)RFDestroyReaderHandle(*phCard);
 			*phCard = 0;
 			rv = SCARD_E_SHARING_VIOLATION;
+			(void)pthread_mutex_unlock(&LockMutex);
 			goto exit;
 		}
 	}
@@ -483,6 +488,8 @@ LONG SCardConnect(/*@unused@*/ SCARDCONTEXT hContext, LPCSTR szReader,
 	 */
 	rv = RFAddReaderHandle(rContext, *phCard);
 
+	(void)pthread_mutex_unlock(&LockMutex);
+
 	if (rv != SCARD_S_SUCCESS)
 	{
 		/*
@@ -880,39 +887,29 @@ LONG SCardDisconnect(SCARDHANDLE hCard, DWORD dwDisposition)
 		 */
 		(void)RFSetReaderEventState(rContext, SCARD_RESET);
 
-		/*
-		 * Currently pcsc-lite keeps the card powered constantly
-		 * unless DISABLE_AUTO_POWER_ON is defined
-		 */
 		dwAtrLen = sizeof(rContext->readerState->cardAtr);
 		if (SCARD_RESET_CARD == dwDisposition)
 			rv = IFDPowerICC(rContext, IFD_RESET,
 				rContext->readerState->cardAtr, &dwAtrLen);
 		else
 		{
+			/* SCARD_UNPOWER_CARD */
 			IFDPowerICC(rContext, IFD_POWER_DOWN, NULL, NULL);
 
-#ifdef DISABLE_AUTO_POWER_ON
 			rContext->powerState = POWER_STATE_UNPOWERED;
 			Log1(PCSC_LOG_DEBUG, "powerState: POWER_STATE_UNPOWERED");
-#else
-			rv = IFDPowerICC(rContext, IFD_POWER_UP,
-				rContext->readerState->cardAtr, &dwAtrLen);
-#endif
 		}
 
 		/* the protocol is unset after a power on */
 		rContext->readerState->cardProtocol = SCARD_PROTOCOL_UNDEFINED;
 
-#ifdef DISABLE_AUTO_POWER_ON
 		if (SCARD_UNPOWER_CARD == dwDisposition)
 		{
-			rContext->readerState->cardAtrLength = 0;
 			if (rv == SCARD_S_SUCCESS)
 				rContext->readerState->readerState = SCARD_PRESENT;
 			else
 			{
-				Log3(PCSC_LOG_ERROR, "Error powering down card: %d 0x%04X",
+				Log3(PCSC_LOG_ERROR, "Error powering down card: %ld 0x%04lX",
 					rv, rv);
 				if (rv == SCARD_W_REMOVED_CARD)
 					rContext->readerState->readerState = SCARD_ABSENT;
@@ -920,10 +917,8 @@ LONG SCardDisconnect(SCARDHANDLE hCard, DWORD dwDisposition)
 					rContext->readerState->readerState =
 						SCARD_PRESENT | SCARD_SWALLOWED;
 			}
-			Log1(PCSC_LOG_INFO, "Skip card power on");
 		}
 		else
-#endif
 		{
 			/*
 			 * Set up the status bit masks on readerState
@@ -1025,16 +1020,8 @@ LONG SCardDisconnect(SCARDHANDLE hCard, DWORD dwDisposition)
 		 * powered */
 		if (POWER_STATE_POWERED <= rContext->powerState)
 		{
-#ifdef DISABLE_AUTO_POWER_ON
-			if (SCARD_RESET_CARD == dwDisposition)
-			{
-				rContext->powerState = POWER_STATE_GRACE_PERIOD;
-				Log1(PCSC_LOG_DEBUG, "powerState: POWER_STATE_GRACE_PERIOD");
-			}
-#else
 			rContext->powerState = POWER_STATE_GRACE_PERIOD;
 			Log1(PCSC_LOG_DEBUG, "powerState: POWER_STATE_GRACE_PERIOD");
-#endif
 		}
 
 		(void)pthread_mutex_unlock(&rContext->powerState_lock);
@@ -1137,6 +1124,18 @@ LONG SCardEndTransaction(SCARDHANDLE hCard, DWORD dwDisposition)
 	if (rv != SCARD_S_SUCCESS)
 		goto exit;
 
+	/*
+	 * Error if another transaction is ongoing and a card action is
+	 * requested
+	 */
+	if ((dwDisposition != SCARD_LEAVE_CARD) && (rContext->hLockId != 0)
+		&& (rContext->hLockId != hCard))
+	{
+		Log1(PCSC_LOG_INFO, "No card reset within a transaction");
+		rv = SCARD_E_SHARING_VIOLATION;
+		goto exit;
+	}
+
 	if (dwDisposition == SCARD_RESET_CARD ||
 		dwDisposition == SCARD_UNPOWER_CARD)
 	{
